import logging
import time
from contextlib import contextmanager
import traceback
import sys

from urllib3.util.retry import Retry
from urllib3.util.timeout import Timeout
from urllib3 import exceptions
from  urllib3.contrib import pyopenssl
import urllib3
import OpenSSL.SSL

from . import error
from .urllib3_custom import CustomPoolManager

urllib3.disable_warnings(exceptions.InsecureRequestWarning)
pyopenssl.inject_into_urllib3()


class Urllib3Transport(object):
    __slots__ = (
        'urllib3_response',
        'op_started',
        'pool',
    )

    def __init__(self, pool=None):
        if pool is None:
            pool = CustomPoolManager()
        self.pool = pool
        self.urllib3_response = None

    def prepare_request(self, req, res):
        pass

    @contextmanager
    def handle_network_error(self):
        try:
            yield
        except exceptions.ReadTimeoutError as ex:
            raise error.OperationTimeoutError(str(ex), ex)
        except exceptions.ConnectTimeoutError as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.ProtocolError as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.SSLError as ex:
            raise error.ConnectError(str(ex), ex)
        except OpenSSL.SSL.Error as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.LocationParseError as ex:
            raise error.MalformedResponseError(str(ex), ex)
        except AttributeError:
            # See https://github.com/urllib3/urllib3/issues/1556
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            found = False
            for frame in frames:
                if (
                        "if host.startswith('[')" in frame.line
                        and 'connectionpool.py' in frame.filename
                    ):
                    found = True
                    break
            if found:
                raise error.MalformedResponseError('Invalid redirect header')
            else:
                raise


    def request(self, req, res):
        self.op_started = time.time()
        if req['resolve']:
            for host, ip in req['resolve'].items():
                self.pool.resolving_cache[host] = ip
        headers = req['headers'] or {}
        if req['content_encoding']:
            if not any(x.lower() == 'accept-encoding' for x in headers):
                headers['accept-encoding'] = req['content_encoding']
        with self.handle_network_error():
            self.urllib3_response = self.pool.urlopen(
                req.method(),
                req['url'],
                headers=headers,
                retries=Retry(
                    total=False,
                    connect=False,
                    read=False,
                    redirect=0,
                    status=None,
                ),
                timeout=Timeout(
                    connect=req['connect_timeout'],
                    read=req['timeout'],
                ),
                preload_content=False,
                decode_content=req['decode_content'],
            )

    def read_with_timeout(self, req, res):
        read_limit = req['content_read_limit']
        chunk_size = 1024
        bytes_read = 0
        while True:
            chunk = self.urllib3_response.read(chunk_size)
            if chunk:
                if read_limit:
                    chunk_limit = min(len(chunk), read_limit - bytes_read)
                else:
                    chunk_limit = len(chunk)
                res._bytes_body.write(chunk[:chunk_limit])
                bytes_read += chunk_limit
                if bytes_read >= read_limit:
                    break
            else:
                break
            if time.time() - self.op_started > req['timeout']:
                raise error.OperationTimeoutError(
                    'Timed out while reading response',
                )

    def prepare_response(self, req, res, err):
        try:
            if err:
                res.error = err
            else:
                try:
                    #headers = {}
                    #for key, val in self.urllib3_response.getheaders().items():
                    #    headers[key.lower()] = val
                    #res._cached['parsed_headers'] = headers
                    res.headers = self.urllib3_response.headers
                    res.status = self.urllib3_response.status

                    if hasattr(self.urllib3_response._connection.sock, 'connection'):
                        res.cert = (
                            self.urllib3_response._connection.sock.connection
                            .get_peer_cert_chain()
                        )

                    with self.handle_network_error():
                        self.read_with_timeout(req, res)
                except error.NetworkError as ex:
                    res.error = err
        finally:
            if self.urllib3_response:
                self.urllib3_response.release_conn()



class Request(object):
    __slots__ = (
        'config',
        'retry_count',
        'meta',
    )

    def __init__(self, retry_count=0, meta=None, **kwargs):
        self.config = self.get_default_config()
        self.setup(**kwargs)
        self.retry_count = retry_count
        self.meta = meta or {}

    def get_default_config(self):
        return {
            'url': None,
            'max_redirects': 0,
            'certinfo': False,
            'extra_valid_status': None,
            'timeout': 10,
            'connect_timeout': 5,
            'resolve': None,
            'raw': False,
            'headers': None,
            'content_encoding': 'gzip,chunked', 
            'decode_content': True,
            'content_read_limit': None,
        }

    def setup(self, **kwargs):
        for key in kwargs:
            assert key in (
                'meta', 'name', 'url',
                'max_redirects', 'certinfo',
                'timeout', 'connect_timeout',
                'resolve', 'raw', 'headers',
                'content_encoding', 'decode_content',
                'content_read_limit',
            ), 'Invalid configuration key: %s' % key
            self.config[key] = kwargs[key]

    def __getitem__(self, key):
        return self.config[key]

    def as_data(self):
        return {
            'config': self.config,
            'retry_count': self.retry_count,
            'meta': self.meta,
        }

    @classmethod
    def from_data(cls, data):
        req = Request()
        for key in ('config', 'retry_count', 'meta'):
            setattr(req, key, data[key])
        return req

    def method(self):
        return 'GET'



import time
import sys
import logging
from threading import Event, Thread, Lock
from queue import Queue, Empty, Full
from urllib.parse import urlsplit 
import re
import json
from urllib.request import urlopen
from traceback import format_exception
from collections import defaultdict

from .util import Pause, debug
#from .loop import MultiCurlLoop
from .network_service import NetworkService
from .stat import Stat
from .request import Request


class Crawler(object):
    _taskgen_sleep_time = 0.01
    dataop_threshold_default = {
        'number': 500,
        'size': None,
        #'time': None,
    }
    dataop_threshold = {}

    def task_generator(self):
        if False:
            yield None

    def __init__(self,
            network_threads=3,
            result_workers=4,
            retry_limit=3,
        ):
        self.taskq = Queue()
        self.taskq_size_limit = max(100, network_threads * 2)
        self.resultq = Queue()
        self.shutdown_event = Event()
        self.network_pause = Pause()
        self.retry_limit = retry_limit
        self.stat = Stat(speed_keys='crawler:request-processed')
        self.fatalq = Queue()
        self.network = NetworkService(
            self.taskq, self.resultq,
            threads=network_threads,
            shutdown_event=self.shutdown_event,
            pause=self.network_pause,
            setup_handler_hook=self.setup_handler_hook,
            stat=self.stat,
        )
        self.result_workers = result_workers
        self._run_started = None

        self.dataopq = {}
        self.dataop_lock = defaultdict(Lock)
        self.dataop_counters = defaultdict(lambda: {
            'number': 0,
            'size': 0,
        })

        self.init_hook()

    def is_dataopq_dump_time(self, name):
        to_dump = False
        try:
            number_th = self.dataop_threshold[name]['number']
        except KeyError:
            number_th = self.dataop_threshold_default['number']
        if number_th:
            if self.dataop_counters[name]['number'] >= number_th:
                    to_dump = True
        if not to_dump:
            try:
                size_th = self.dataop_threshold[name]['size']
            except KeyError:
                size_th = self.dataop_threshold_default['size']
            if size_th:
                if self.dataop_counters[name]['size'] >= size_th:
                    to_dump = True
        return to_dump

    def enq_dataop(self, name, op, size=None, force_dump=False):
        self.dataop_lock[name].acquire()
        released = False
        try:
            self.dataopq.setdefault(name, [])
            if op:
                self.dataopq[name].append(op)
                self.dataop_counters[name]['number'] += 1
                if size:
                    self.dataop_counters[name]['size'] += size

            if force_dump or self.is_dataopq_dump_time(name):
            #if not len(self.dataopq[name]) % 500:
            #    logging.debug('Size of %s docs: %d' % (len(self.dataopq[name]), size))
                logging.debug('Dumping data ops')
                ops = self.dataopq[name]
                self.dataopq[name] = []
                self.dataop_counters[name]['number'] = 0
                self.dataop_counters[name]['size'] = 0
                self.dataop_lock[name].release()
                released = True
                func = getattr(self, 'dataop_handler_%s' % name)
                func(ops)
        finally:
            if not released:
                self.dataop_lock[name].release()


    def init_hook(self):
        pass

    def setup_handler_hook(self, transport, req):
        pass

    def submit_task(self, req):
        self.submit_task_hook(req)
        self.taskq.put(req)

    def submit_task_hook(self, req):
        pass

    def thread_task_generator(self):
        try:
            #sleep_time = 0.01
            for item in self.task_generator():
                while item:
                    if self.shutdown_event.is_set():
                        return
                    if self.taskq.qsize() >= self.taskq_size_limit:
                        time.sleep(self._taskgen_sleep_time)
                    else:
                        self.submit_task(item)
                        item = None
        except Exception as ex:
            self.fatalq.put(sys.exc_info())

    def is_result_ok(self, req, res):
        if res.error:
            return False
        elif (
                0 < res.status < 400
                or res.status == 404
                or (
                    req.config['extra_valid_status']
                    and res.status in req.config['extra_valid_status']
                )
            ):
            return True
        else:
            return False

    def thread_result_processor(self, pause):
        try:
            while not self.shutdown_event.is_set():
                if pause.pause_event.is_set():
                    pause.process_pause()
                try:
                    result = self.resultq.get(True, 0.1)
                except Empty:
                    pass
                else:
                    self.stat.inc('crawler:request-processed')
                    if (
                            result['request']['raw']
                            or self.is_result_ok(
                                result['request'],
                                result['response'],
                            )
                        ):
                        self.process_ok_result(result)
                    else:
                        self.process_fail_result(result)
        except Exception as ex:
            self.fatalq.put(sys.exc_info())

    def thread_fatalq_processor(self):
        while not self.shutdown_event.is_set():
            try:
                etype, evalue, tb = self.fatalq.get(True, 0.1)
            except Empty:
                pass
            else:
                self.shutdown_event.set()
                logging.error('Fatal exception')
                logging.error(''.join(format_exception(etype, evalue, tb)))

    def thread_manager(self, th_task_gen, pauses):
        try:
            th_task_gen.join()

            def system_is_busy():
                return (
                    self.taskq.qsize()
                    or self.resultq.qsize()
                    or len(self.network.active_handlers)
                )

            while not self.shutdown_event.is_set():
                while system_is_busy():
                    if self.shutdown_event.is_set():
                        return
                    time.sleep(0.1)

                for pause in pauses:
                    pause.pause_event.set()
                ok = True
                for pause in pauses:
                    if not pause.is_paused.wait(0.1):
                        ok = False
                        break
                if not ok:
                    for pause in pauses:
                        pause.pause_event.clear()
                        pause.resume_event.set()
                else:
                    if not system_is_busy():
                        for pause in pauses:
                            pause.pause_event.clear()
                            pause.resume_event.set()
                        self.shutdown_event.set()
                        return
        except Exception as ex:
            self.fatalq.put(sys.exc_info())

    def process_ok_result(self, result):
        self.stat.inc('crawler:request-ok')
        name = result['request'].config['name']
        handler = getattr(self, 'handler_%s' % name)
        handler_result = handler(
            result['request'],
            result['response'],
        )
        try:
            iter(handler_result)
        except TypeError:
            pass
        else:
            for item in handler_result:
                print('Processing handler result: %s' % item)

    def process_fail_result(self, result):
        self.stat.inc('crawler:request-fail')
        req = result['request']

        if result['response'].error:
            self.stat.inc('network-error:%s' % result['response'].error.get_tag())
        if result['response'].status:
            self.stat.inc('http:status-%s' % result['response'].status)

        if req.retry_count + 1 <= self.retry_limit:
            self.stat.inc('crawler:request-retry')
            req.retry_count += 1
            self.submit_task(req)
        else:
            self.stat.inc('crawler:request-rejected')
            name = result['request'].config['name']
            handler = getattr(self, 'rejected_%s' % name, None)
            if handler is None:
                handler = self.default_rejected_handler
            handler(result['request'], result['response'])

    def default_rejected_handler(self, req, res):
        pass

    def prepare(self):
        pass

    def thread_stat(self):
        try:
            while not self.shutdown_event.is_set():
                stat = []
                now = time.time()
                for hdl in self.network.active_handlers:
                    stat.append((hdl, self.network.registry[hdl]['start']))
                with open('var/crawler.stat', 'w') as out:
                    for hdl, start in list(sorted(stat, key=lambda x: (x[1] or now), reverse=True)):
                        req = self.network.registry[hdl]['request']
                        out.write('%.2f - [#%s] - %s\n' % (
                            (now - start) if start else 0,
                            req.retry_count if req else 'NA',
                            urlsplit(req['url']).netloc if req else 'NA',
                        ))
                    out.write('Active handlers: %d\n' % len(self.network.active_handlers))
                    out.write('Idle handlers: %d\n' % len(self.network.idle_handlers))
                    out.write('Taskq size: %d\n' % self.taskq.qsize())
                    out.write('Resultq size: %d\n' % self.resultq.qsize())

                total = 0
                count = 0
                for hdl, start in stat:
                    if start:
                        total += (now - start)
                        count += 1
                logging.debug('Median handler time: %.2f' % ((total / count) if count else 0))

                self.shutdown_event.wait(3)
        except Exception as ex:
            self.fatalq.put(sys.exc_info())


    def thread_network(self):
        try:
            self.network.run()
        except Exception as ex:
            self.fatalq.put(sys.exc_info())

    def shutdown(self):
        for name in self.dataopq.keys():
            self.enq_dataop(name, None, force_dump=True)
        self.shutdown_hook()

    def shutdown_hook(self):
        pass

    def run(self):
        self.prepare()
        try:
            self._run_started = time.time()

            th_fatalq_proc = Thread(target=self.thread_fatalq_processor)
            th_fatalq_proc.start()

            th_task_gen = Thread(target=self.thread_task_generator)
            th_task_gen.start()


            th_stat = Thread(target=self.thread_stat)
            th_stat.start()


            pauses = [self.network_pause]
            result_workers = []
            for _ in range(self.result_workers):
                pause = Pause()
                th = Thread(
                    target=self.thread_result_processor,
                    args=[pause],
                ) 
                pauses.append(pause)
                th.start()
                result_workers.append(th)

            th_manager = Thread(
                target=self.thread_manager,
                args=[th_task_gen, pauses],
            )
            th_manager.start()


            th_network = Thread(
                target=self.thread_network,
            )
            th_network.start()

            #def force_shutdown():
            #    if not self.shutdown_event.wait(5):
            #        print('Forcing shutdown')
            #        self.shutdown_event.set()

            #Thread(target=force_shutdown).start()

            try:
                th_manager.join()
                th_fatalq_proc.join()
                th_task_gen.join()
                th_stat.join()
                [x.join() for x in result_workers]
            except KeyboardInterrupt:
                self.shutdown_event.set()
        finally:
            self.shutdown()



from gevent import monkey
monkey.patch_all()#thread=False)

try:
    import grpc
except ImportError:
    pass
else:
    if grpc.__version__ != '1.18.0':
        raise Exception('grpcio version must be 1.18.0')
    from grpc.experimental import gevent
    gevent.init_gevent()

from .session import Session
from .request import Request
from .data import Data
from .response import Response
from .crawler import Crawler
from .transport import Urllib3Transport
from .task_generator import TaskGenerator



"""
https://curl.haxx.se/libcurl/c/libcurl-errors.html
"""
#import pycurl
#
#from .pycurl_error import PYCURL_ERRNO_TAG


class IowebError(Exception):
    pass


class NetworkError(IowebError):
    """
    NetworkError instance stores original transport-specific error
    in `transport_error` attribute.
    """
    def __init__(self, *args):
        self.errmsg = None
        self.transport_error = None
        if len(args) > 0:
            self.errmsg = args[0]
        if len(args) > 1:
            self.transport_error = args[1]

    def get_tag(self):
        #if isinstance(self.transport_error, pycurl.error):
        #    errno = self.transport_error.args[0]
        #    errmsg = self.transport_error.args[1]
        #    if (
        #            errno == 28 and (
        #                'Connection time-out' in errmsg
        #                or 'Connection timed out' in errmsg
        #            )
        #        ):
        #        return 'connection-timed-out'
        #    else:
        #        return PYCURL_ERRNO_TAG[errno]
        #else:
        return self.transport_error.__class__.__name__.lower()


class DataWriteError(NetworkError):
    """
    CURLE_WRITE_ERROR (23)

    An error occurred when writing received data to a local file,
    or an error was returned to libcurl from a write callback.
    """


class OperationTimeoutError(NetworkError):
    """
    CURLE_OPERATION_TIMEDOUT (28)

    Operation timeout. The specified time-out period was reached
    according to the conditions.
    """


class ConnectError(NetworkError):
    """
    CURLE_COULDNT_CONNECT (7)

    Failed to connect() to host or proxy.
    """


class AuthError(NetworkError):
    """
    CURLE_LOGIN_DENIED (67)

    The remote server denied curl to login (Added in 7.13.1)
    """


class TooManyRedirectsError(NetworkError):
    """
    CURLE_TOO_MANY_REDIRECTS (47)

    Too many redirects. When following redirects, libcurl hit the
    maximum amount. Set your limit with CURLOPT_MAXREDIRS.
    """


class ResolveHostError(NetworkError):
    """
    CURLE_COULDNT_RESOLVE_HOST (6)

    Couldn't resolve host. The given remote host was not resolved.
    """


class InvalidUrlError(NetworkError):
    """
    CURLE_URL_MALFORMAT (3)

    The URL was not properly formatted.
    """


class MalformedResponseError(NetworkError):
    """
    Raised when parser fails to parse response HTTP headers
    """


ERRNO_CLASS_MAPPING = {
    3: InvalidUrlError,
    6: ResolveHostError,
    7: ConnectError,
    23: DataWriteError,
    28: OperationTimeoutError,
    47: TooManyRedirectsError,
    67: AuthError,
}


#def build_network_error(errno, errmsg):
#    cls = ERRNO_CLASS_MAPPING.get(errno, NetworkError)
#    return cls(errmsg, pycurl.error(errno, errmsg))



"""
Credit: https://github.com/Roadmaster/forcediphttpsadapter/blob/master/forcediphttpsadapter/adapters.py
"""
import logging

from urllib3.util import connection
from urllib3.connectionpool import HTTPSConnectionPool, HTTPConnectionPool
from urllib3.connection import HTTPSConnection, DummyConnection, HTTPConnection
from urllib3 import PoolManager
from urllib3.poolmanager import SSL_KEYWORDS
from socket import error as SocketError, timeout as SocketTimeout
from urllib3.exceptions import (
    ConnectTimeoutError,
    NewConnectionError,
)
from cachetools import TTLCache

log = logging.getLogger(__name__)


class CustomHttpsConnection(HTTPSConnection, object):
    def __init__(self, *args, **kwargs):
        # custom: begins
        self.custom_ip = kwargs.pop('custom_ip', None)
        # custom: ends
        super(CustomHttpsConnection, self).__init__(*args, **kwargs)

    def _new_conn(self):
        extra_kw = {}
        if self.source_address:
            extra_kw['source_address'] = self.source_address

        if self.socket_options:
            extra_kw['socket_options'] = self.socket_options

        custom_host = self.custom_ip if self.custom_ip else self._dns_host
        #if self.custom_ip:
        #    logging.debug('Using custom ip %s for host %s' % (self.custom_ip, self._dns_host))

        try:
            conn = connection.create_connection(
                (custom_host, self.port), self.timeout, **extra_kw)

        except SocketTimeout as e:
            raise ConnectTimeoutError(
                self, "Connection to %s timed out. (connect timeout=%s)" %
                (self.host, self.timeout))

        except SocketError as e:
            raise NewConnectionError(
                self, "Failed to establish a new connection: %s" % e)

        return conn


class CustomHttpConnection(HTTPConnection, object):
    def __init__(self, *args, **kwargs):
        # custom: begins
        self.custom_ip = kwargs.pop('custom_ip', None)
        # custom: ends
        super(CustomHttpConnection, self).__init__(*args, **kwargs)

    def _new_conn(self):
        extra_kw = {}
        if self.source_address:
            extra_kw['source_address'] = self.source_address

        if self.socket_options:
            extra_kw['socket_options'] = self.socket_options

        custom_host = self.custom_ip if self.custom_ip else self._dns_host
        #if self.custom_ip:
        #    logging.debug('Using custom ip %s for host %s' % (self.custom_ip, self._dns_host))

        try:
            conn = connection.create_connection(
                (custom_host, self.port), self.timeout, **extra_kw)

        except SocketTimeout as e:
            raise ConnectTimeoutError(
                self, "Connection to %s timed out. (connect timeout=%s)" %
                (self.host, self.timeout))

        except SocketError as e:
            raise NewConnectionError(
                self, "Failed to establish a new connection: %s" % e)

        return conn


class CustomHttpConnectionPool(HTTPConnectionPool):
    def __init__(self, *args, **kwargs):
        # custom: begins
        self.custom_ip = kwargs.pop('custom_ip', None)
        # custom: ends
        super(CustomHttpConnectionPool, self).__init__(*args, **kwargs)

    def _new_conn(self):
        self.num_connections += 1
        log.debug("Starting new HTTP connection (%d): %s:%s",
                  self.num_connections, self.host, self.port or "80")


        # custom: begins
        self.conn_kw = getattr(self, 'conn_kw', {})
        self.conn_kw['custom_ip'] = self.custom_ip
        # custom: ends

        conn = CustomHttpConnection(
            host=self.host, port=self.port,
            timeout=self.timeout.connect_timeout,
            strict=self.strict, **self.conn_kw
        )
        return conn

class CustomHttpsConnectionPool(HTTPSConnectionPool):
    def __init__(self, *args, **kwargs):
        # custom: begins
        self.custom_ip = kwargs.pop('custom_ip', None)
        # custom: ends
        super(CustomHttpsConnectionPool, self).__init__(*args, **kwargs)

    def _new_conn(self):
        self.num_connections += 1
        log.debug("Starting new HTTPS connection (%d): %s:%s",
                  self.num_connections, self.host, self.port or "443")

        if not self.ConnectionCls or self.ConnectionCls is DummyConnection:
            raise SSLError("Can't connect to HTTPS URL because the SSL "
                           "module is not available.")
        actual_host = self.host
        actual_port = self.port
        if self.proxy is not None:
            actual_host = self.proxy.host
            actual_port = self.proxy.port

        # custom: begins
        self.conn_kw = getattr(self, 'conn_kw', {})
        self.conn_kw['custom_ip'] = self.custom_ip
        # custom: ends

        conn = CustomHttpsConnection(
            host=actual_host,
            port=actual_port,
            timeout=self.timeout.connect_timeout,
            strict=self.strict,
            # custom: begins
            #cert_file=self.cert_file,
            #key_file=self.key_file,
            #key_password=self.key_password,
            # custom: ends
            **self.conn_kw
        )
        return self._prepare_conn(conn)

    # custom: begins
    def __str__(self):
        return ('%s(host=%r, port=%r, custom_ip=%s)' % (
            type(self).__name__,
            self.host,
            self.port,
            self.custom_ip,
        ))
    # custom: ends


class CustomPoolManager(PoolManager):

    def __init__(self, *args, **kwargs):
        super(CustomPoolManager, self).__init__(*args, **kwargs)
        # custom: begins
        self.resolving_cache = TTLCache(maxsize=10000, ttl=(60 * 60))
        self.pool_classes_by_scheme = {
            'http': CustomHttpConnectionPool,
            'https': CustomHttpsConnectionPool,
        }
        # custom: ends

    # custom: begins
    def get_custom_ip(self, host):
        return self.resolving_cache.get(host, None)
    # custom: ends

    def _new_pool(self, scheme, host, port, request_context=None):
        pool_cls = self.pool_classes_by_scheme[scheme]
        if request_context is None:
            request_context = self.connection_pool_kw.copy()

        # Although the context has everything necessary to create the pool,
        # this function has historically only used the scheme, host, and port
        # in the positional args. When an API change is acceptable these can
        # be removed.
        for key in ('scheme', 'host', 'port'):
            request_context.pop(key, None)

        if scheme == 'http':
            for kw in SSL_KEYWORDS:
                request_context.pop(kw, None)

        # custom: begins
        custom_ip = self.get_custom_ip(host)
        if custom_ip:
            request_context['custom_ip'] = self.resolving_cache[host]
        # custom: ends

        return pool_cls(host, port, **request_context)



from collections import defaultdict, deque
import time
import logging


class Stat(object):
    default_aliases = {
        'crawler:request-processed': 'req',
        'crawler:request-ok': 'req-ok',
        'crawler:request-retry': 'req-retry',
        'crawler:request-fail': 'req-fail',
        'crawler:request-rejected': 'req-rejected',
    }
    ignore_prefixes = (
        'http:',
        'network-error:',
    )
    def __init__(
            self,
            speed_keys=None,
            logging_interval=3,
            snapshot_interval=3,
            num_snapshots=20,
            aliases=None,
        ):
        if speed_keys is None:
            speed_keys = []
        elif isinstance(speed_keys, str):
            speed_keys = [speed_keys]
        self.speed_keys = speed_keys
        self.counters = defaultdict(int)
        self.snapshots = deque(maxlen=num_snapshots)
        self.snapshot_time = 0
        self.snapshot_interval = snapshot_interval
        self.logging_time = 0
        self.logging_interval = logging_interval
        self.aliases = self.default_aliases
        if aliases:
            self.aliases.update(aliases)

    def inc(self, key, count=1):
        self.counters[key] += count
        now = time.time()
        if now - self.snapshot_time > self.snapshot_interval:
            shot = {}
            for key in self.speed_keys:
                shot[key] = self.counters[key]
            self.snapshots.append((now, shot))
            self.snapshot_time = now
        if now - self.logging_time > self.logging_interval:
            rps_items = []
            if len(self.snapshots) > 1:
                last_shot = self.snapshots[-1] if self.snapshots else {}
                first_shot = self.snapshots[0] if self.snapshots else {}
                for key in self.speed_keys:
                    rps_items.append((
                        key,
                        (last_shot[1].get(key, 0) - first_shot[1].get(key, 0))
                        / (last_shot[0] - first_shot[0])
                    ))
            rps_str = ', '.join(
                '%s: %.1f' % (self.aliases.get(key, key), val) for key, val in rps_items
            )
            keys = self.counters.keys()
            counter_str = ', '.join(
                '%s: %s' % (self.aliases.get(x, x), self.counters[x])
                for x in keys
                if not x.startswith(self.ignore_prefixes)
            )
            logging.debug('Stat:%s [%s]', rps_str, counter_str)
            self.logging_time = now
